(* document.sml
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *)

structure Document =
  struct

  (* The following type describes a "canonical" tagging annotation.
   * The idea is that successive applications of such taggings can
   * always be collapsed into a single one.  Tags are listed in
   * outside-in order (e.g., <color><tt> ... </tt></color>).
   * Invariant: s implies not em.
   *)
    type tagging = {
	color : int,
	sz : int,
	font : word	(* U@0..1, EM@2, S@3, TT@4, I@5, B@6, PL@7 *)
      }

  (* A document "ignores" or "tolerates" a tag if no matter
   * whether or not that tag is added as an outside wrapper
   * it won't change the meaning of the document.
   *)
    type ignore = {
	  color : bool,
	  sz : bool,
	  font : word	(* same bit layout as tagging *)
	}

    datatype doc
      = SP of {ign : ignore}
      | TEXT of {body : string, ign : ignore}
      | LIST of {body : doc list, ign : ignore}
      | ELEM of {
	  tags : tagging,
	  content : doc,
	  ign : ignore
	}

    fun ignOf (SP{ign, ...}) = ign
      | ignOf (TEXT{ign, ...}) = ign
      | ignOf (LIST{ign, ...}) = ign
      | ignOf (ELEM{ign, ...}) = ign

  end
