(* -*- sml -*-
 * em-opt.sml
 *
 * (C) 2001 Lucent Technologies, Bell Labs
 *)
structure EMOpt : sig
    (* Improve the number of EM tags in the document by "bubbling them
     * up the tree". *)
    val emopt : Document.doc -> Document.doc
end = struct

    structure D = Document
    structure II = FastIInfer
    structure U = OptUtil

#include "bits.ins"

    (* YES : "I need an EM tag around me"
     * NO  : "I don't want an EM tag around me"
     * NEUTRAL : "Don't put an EM tag on me, but also don't count me
     *            as a NO."
     *)

    datatype flip = YES | NEUTRAL | NO

    fun emopt doc = let

	fun push (x as (D.SP _ | D.TEXT _ | D.LIST _)) =
	    D.ELEM { content = x,
		     tags = { color = ~1, sz = ~1, font = em_bit },
		     ign = D.ignOf x }
	  | push (x as D.ELEM { tags, content, ign }) = let
		val font = #font tags
	    in
		if testBit s_bit font orelse testBit pl_bit font then x
		else
		    D.ELEM { tags = { color = #color tags,
				      sz = #sz tags,
				      font = flipBit em_bit (#font tags) },
			     content = content,
			     ign = ign }
	    end

	fun pushwhen (when: flip) (d, em) =
	    if when = em then push d else d

	fun em (x as (D.SP _ | D.TEXT _)) = (x, NO)
	  | em (D.ELEM { tags, content, ign }) = let
		val font = #font tags
		val (content2, embubble) = em content
	    in
		if testBit s_bit font then
		    (D.ELEM { tags = tags, content = content2, ign = ign },
		     NEUTRAL)
		else if testBit pl_bit font then
		    if embubble = YES then
			(D.ELEM { tags = { color = #color tags,
					   sz = #sz tags,
					   font = flipBit em_bit font },
				  content = content2,
				  ign = ign },
			 NEUTRAL)
		    else
			(D.ELEM { tags = tags, content = content2, ign = ign },
			 NEUTRAL)
		else let
			val myem = testBit em_bit font
			val newem = (embubble = YES) <> myem
			val tags2 = { color = #color tags,
				      sz = #sz tags,
				      font = clearBit em_bit font }
		    in
			(if U.isNop tags2 then content2
			 else D.ELEM { tags = tags2, content = content2,
				       ign = ign },
			 if newem then YES else NO)
		    end
	    end
	  | em (D.LIST { body, ign }) = let
		val body_ems = map em body
		fun isYES (_, YES) = true
		  | isYES _ = false
		fun isNO (_, NO) = true
		  | isNO _ = false
		val yescount = length (List.filter isYES body_ems)
		val nocount = length (List.filter isNO body_ems)
		fun finish (when, bubble) = let
		    val body2 = map (pushwhen when) body_ems
		    val ign2 = II.i_list (map D.ignOf body2)
		in
		    (D.LIST { body = body2, ign = ign2 }, bubble)
		end
	    in
		case Int.compare (yescount, nocount) of
		    GREATER => finish (NO, YES)
		  | EQUAL => finish (YES, NEUTRAL)
		  | LESS => finish (YES, NO)
	    end
    in
	pushwhen YES (em doc)
    end
end
