#include <stdio.h>
#include "collapse.h"

#define FA_SZ 128

unsigned char fa[FA_SZ][7];

void error(char *s) {
    fprintf(stderr, "error: %s\n", s);
    exit(1);
}

typedef unsigned char uchar;

typedef struct {
    int font;
    uchar szcol;
} tagging;

void tcollapse(tagging *t1, tagging *t2, tagging *res) {
    int col1 = CL(t1->szcol);
    int sz1 = SZ(t1->szcol);
    int col2 = CL(t2->szcol);
    int sz2 = SZ(t2->szcol);
    int szcol = MK_SZCOL(sz1 == -1 ? sz2 : sz1,col1 == -1 ? col2 : col1);
    int u, b, i, s, tt, pl, r, em;
   
    if(t2->font & PL) {
	*res = *t2;
	res->szcol = szcol;
	return;
    }
    u = (t1->font & U) + (t2->font & U);
    if(u>3) u=3;
    b = (t1->font & B) | (t2->font & B);
    i = (t1->font & I) | (t2->font & I);
    tt = (t1->font & TT) | (t2->font & TT);
    s = (t1->font & S) | (t2->font & S);
    em = (!s && ((t1->font & EM) != (t2->font & EM))) << (EMb+1);
    pl = (!((u==3) && b && i && tt && s) && (t1->font & PL)) << (PLb+1);
    res->szcol = szcol;
    res->font = u|b|i|tt|s|em|pl;
}

#define MAX(a,b) ((a) > (b) ? (a) : (b))

void igelem(tagging *t, tagging *in, tagging *res) {
    int tpl = (t->font & PL) != 0;
    int u,b,i,tt,s,em,pl;

    u = tpl ? 0 : MAX((in->font & U) - (t->font & U), 0);
    b = (tpl || (t->font & B) || (in->font & B)) << (Bb+1);
    i = (tpl || (t->font & I) || (in->font & I)) << (Ib+1);
    tt = (tpl || (t->font & TT) || (in->font & TT)) << (TTb+1);
    s = (tpl || (t->font & S) || (in->font & S)) << (Sb+1);
    em = (tpl || s || (in->font & EM)) << (EMb+1);
    pl = (tpl || ((in->font & PL) != 0)) << (PLb+1);
    // This was wrong:
    // pl = (!b && !i && !tt && !em && u == 3) << (PLb+1);
    res->font = u|b|i|tt|s|em|pl;
}

void mkigelemtab() {
    tagging t1, t2, res;
    int i, j;
    unsigned char c;
 
    for(i = 0; i < 256; i++) {
//	printf("igelemtab[%d] = \"",i);
	for(j = 0; j < 256; j++) {
	    t1.font = i; 
	    t2.font = j;
	    igelem(&t1,&t2,&res);
//	    printf("\\%03d", res.font);
	    c = res.font;
	    write(1,&c,1);
	}
//	printf("\"\n");
    }
}
    
void mkcollapsetab() {
    tagging t1, t2, res;
    int i, j;
    char c;
 
    for(i = 0; i < 256; i++) {
//	printf("tcollapse[%d] = \"",i);
	for(j = 0; j < 256; j++) {
	    t1.font = i; 
	    t2.font = j;
	    tcollapse(&t1,&t2,&res);
//	    printf("\\%03d", res.font);
	    c = res.font;
	    write(1,&c,1);
	}
//	printf("\"\n");
    }
}

	    
	
int main(int argc, char *argv[]) {
    
    if(argc != 2) {
	fprintf(stderr, "usage: %s [0|1]\n0 generates igelemtab, 1 generates collapsetab\n", argv[0]);
	exit(1);
    }

    if(atoi(argv[1]) == 1)
	mkcollapsetab();
    else
	mkigelemtab();
    return 0;
}


