#include <stdio.h>
#include <string.h>
#include "fa.h"

int verbose = 0;
#ifndef DECOR
int emitdecor = 0;
#else
int emitdecor = 1;
#endif

#define ONE_MB (1024*1024)

char f[6*ONE_MB];
int f_sz;
char g[6*ONE_MB];
int g_sz;

#define R_SZ 1024

typedef struct {
    int color;
    int size;
    int state;
    char *s;
} context;

typedef struct {
    char attr;
    int start;
    int len;
    int kids;
    int parent;
    int next;
    int dead;
    context ctxt;
} node;

node nodes[ONE_MB];
int curnode = 0;

char *ttab[256];

void init() {
    ttab['I'] = "I";
    ttab['B'] = "B";
    ttab['S'] = "S";
    ttab['U'] = "U";
    ttab['E'] = "EM";
    ttab['T'] = "TT";
    ttab['P'] = "PL";
    ttab['R'] = "ROOT";
    ttab['0'] = "0";
    ttab['1'] = "1";
    ttab['2'] = "2";
    ttab['3'] = "3";
    ttab['4'] = "4";
    ttab['5'] = "5";
    ttab['6'] = "6";
    ttab['7'] = "7";
    ttab['8'] = "8";
    ttab['9'] = "9";
    ttab['r'] = "r";
    ttab['g'] = "g";
    ttab['b'] = "b";
    ttab['c'] = "c";
    ttab['m'] = "m";
    ttab['y'] = "y";
    ttab['k'] = "k";
    ttab['w'] = "w";
}



int mknode(char attr, int start, int parent, context *ct) {
    nodes[curnode].kids = -1;
    nodes[curnode].attr = attr;
    nodes[curnode].parent = parent;
    nodes[curnode].next = -1;
    nodes[curnode].start = start;
    nodes[curnode].ctxt = *ct;
    nodes[curnode].dead = 0;
    return curnode++;
}

#define STACK_SZ ONE_MB
context stack[STACK_SZ];
int sp = 0;

int readfile(int fid) {
    int n;

    while((n = read(fid, f+f_sz, R_SZ)) > 0)
	f_sz += n;
    return n < 0 ? 0 : 1;
}

char lastemit = '\0';
int emitpos = 0;

void emit(char c, context *f) {
    lastemit = c;
    g[emitpos++] = c;
    if(emitdecor) {
	putchar(c);
	printf("<%c,%d,%x>\n", f->color, f->size, f->state);
    }
}

void push(context *f) {
    if(sp >= STACK_SZ) {
	fprintf(stderr, "stack overflow\n");
	exit(1);
    }
	
    stack[sp].color = f->color;
    stack[sp].size = f->size;
    stack[sp].state = f->state;
    stack[sp].s = f->s;
    sp++;
}

void pop(context *frame) {
    if(sp <  1) {
	fprintf(stderr, "stack underflow\n");
	exit(1);
    }
    *frame = stack[--sp];
}

void dumpparse(int n) {
    int i;

    if(n>=0) {
	printf("{%d,%c;%d:%d}(", n, nodes[n].attr, nodes[n].start, nodes[n].len);
	for(i=nodes[n].kids; i != -1; i=nodes[i].next)
	    dumpparse(i);
	printf(")");
    }
}	

    

int revlist(int n) {
    int q;
    int p = -1;

    while(n != -1) {
	q = nodes[n].next;
	nodes[n].next = p;
	p = n;
	n = q;
    }
    return p;
}

void revparse(int n) {
    int i;

    if(n>=0) {
	nodes[n].kids = revlist(nodes[n].kids);
	for(i=nodes[n].kids; i != -1; i=nodes[i].next)
	    revparse(i);
    }
}

int emitparse(int n, int curpos) {
    int i;
    if(n >= 0) {
	for(i = curpos; i < nodes[n].start; i++)
	    putchar(g[i]);
	curpos = nodes[n].start;
	if(!nodes[n].dead && nodes[n].attr != 'R')
	    printf("<%s>",ttab[nodes[n].attr]);
	for(i = nodes[n].kids; i != -1; i = nodes[i].next)
	    curpos = emitparse(i,curpos);
	for(i = curpos; i < nodes[n].start+nodes[n].len; i++)
	    putchar(g[i]);
	if(!nodes[n].dead && nodes[n].attr != 'R')
	    printf("</%s>",ttab[nodes[n].attr]);
	curpos=i;
    }
    return curpos;
}

int finishnode(int n) {
    nodes[n].len = emitpos - nodes[n].start;
    return nodes[n].parent;
}

int startnode(char attr,int n,context *ct) {
    int node;
    node = mknode(attr,emitpos,n,ct);
    if(nodes[n].kids != -1) {
	nodes[node].next = nodes[n].kids;
    }
    nodes[n].kids = node;
    return node;
}

int cmpCtxt(context *c1, context *c2) {
    return c1->state == c2->state && 
	   c1->size == c2->size && 
	   c1->color == c2->color;
}

void optDead(int n) {
    context *cp;
    int i;

    cp = &nodes[n].ctxt;
    for(i = nodes[n].kids; i != -1; i = nodes[i].next) {
	if(!nodes[i].len)
	    nodes[i].dead=1;
	if(cmpCtxt(cp,&nodes[i].ctxt))
	    nodes[i].dead=1;
	optDead(i);
    }
}

int lapp(int l1, int l2) {
    int i;
    if(l1<0)
	return l2;
    for(i = l1; i != -1; i = nodes[i].next)
	if(nodes[i].next == -1)
	    break;
    nodes[i].next = l2;
    return l1;
}

    
void optCoalese(int n) {
    int i, j, previ = -1;

    for(i = nodes[n].kids; i != -1; i = nodes[i].next) {
	if(previ >= 0 &&
	   nodes[i].ctxt.state == nodes[previ].ctxt.state &&
	   nodes[i].ctxt.color == nodes[previ].ctxt.color &&
	   nodes[i].ctxt.size == nodes[previ].ctxt.size &&
	   nodes[previ].start+nodes[previ].len == nodes[i].start) {
	    nodes[previ].len += nodes[i].len;
	    nodes[previ].kids = lapp(nodes[previ].kids,nodes[i].kids);
	    for(j = nodes[i].kids; j != -1; j = nodes[j].next)
		nodes[j].parent = previ;
	    nodes[previ].next = nodes[i].next;
	}
	optCoalese(i);
	previ = i;
    }
}

void optSEM(int n,int dropEM) {
    int i;

    if(nodes[n].ctxt.state & S)
	dropEM=1;
    for(i = nodes[n].kids; i != -1; i = nodes[i].next) {
	if(dropEM && (nodes[n].attr & EM))
	    nodes[i].dead=1;
	optSEM(i,dropEM);
    }
}

void optEM(int n) {
    int k,i;

    if((nodes[n].attr & EM) && 
       ((k = nodes[n].kids) != -1) &&
       (nodes[k].attr & EM) &&
       (nodes[k].start == nodes[n].start))
	nodes[k].dead=1;
    for(i=nodes[n].kids; i != -1; i = nodes[i].next)
	optEM(i);
}

int eq(int x,int y) {
    return x == y;
    if(x == -1 || y == -1 || x == y)
	return 1;
    return 0;
}

int contextualize(char *p, int n) {
    int i;
    int intag = 0, inclose = 0;
    context frame;
    context cur;
    context space, last_space;
    int this;

    cur.color = 'x';
    cur.size = -1;
    cur.state = 0;
    cur.s = "<initial>";

    last_space = cur;

    this = mknode('R', emitpos, -1,&cur);
    for(i = 0; i < n; i++) {
	switch(p[i]) {
	  case '<':
	    if(intag) {
		fprintf(stderr, "malformed open-tag at position %d\n", i);
		return 0;
	    }
	    intag=1;
	    break;
	  case '>': 
	    if(intag != 2) {
		fprintf(stderr, "malformed close-tag at position %d\n", i);
		return 0;
	    }
	    intag = inclose = 0;
	    break;
	  case 'B':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
		if(inclose) {
		    if(verbose) printf("</bold>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    if(verbose) printf("<bold>");
		    push(&cur);
		    cur.state = cur.state | B;;
		    cur.s = "<bold>";
		    this = startnode('B',this,&cur);
		}
	    }
	    break;
	  case 'I':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
		if(inclose) {
		    if(verbose) printf("</italic>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    if(verbose) printf("<italic>");
		    push(&cur);
		    cur.state = cur.state | I;
		    cur.s = "<italic>";
		    this = startnode('I',this,&cur);
		}
	    }
	    break;
	  case 'S':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
		if(inclose) {
		    if(verbose) printf("</strong>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    if(verbose) printf("<strong>");
		    push(&cur);
		    cur.state = (cur.state & ~EM) | S;
		    cur.s = "<strong>";
		    this = startnode('S',this,&cur);
		}
	    }
	    break;
	  case 'U':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
		if(inclose) {
		    if(verbose) printf("</underline>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    int u;

		    if(verbose) printf("<underline>");
		    push(&cur);
		    u = GET_U(cur.state);
		    if(u < 3) {
			u++;
			cur.state = (cur.state & ~U) | u;
		    }
		    cur.s = "<underline>";
		    this = startnode('U',this,&cur);
		}
	    }
	    break;
	  case 'E':
	  case 'P':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
	    }
	    break;
	  case 'T':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
	    } else if(intag == 2 && p[i-1] == 'T') {
		if(inclose) {
		    if(verbose) printf("</typewriter>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    if(verbose) printf("<typewriter>");
		    push(&cur);
		    cur.state = cur.state | TT;
		    cur.s = "<typewriter>";
		    this = startnode('T',this,&cur);
		}
	    } else {
		fprintf(stderr, "malformed typewriter at position %d\n", i);
		return 0;
	    } 
	    break;
	  case 'L':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 2 && p[i-1] == 'P') {
		if(inclose) {
		    if(verbose) printf("</plain>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    if(verbose) printf("<plain>");
		    push(&cur);
		    cur.state = 0;
		    cur.s = "<plain>";
		    this = startnode('P',this,&cur);
		}
	    } else {
		fprintf(stderr, "malformed plain at position %d\n", i);
		return 0;
	    } 
	    break;
	  case 'M':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 2 && p[i-1] == 'E') {
		if(inclose) {
		    if(verbose) printf("</emphasis>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.state = frame.state;
		} else {
		    if(verbose) printf("<emphasis>");
		    push(&cur);
		    cur.state = !(cur.state & S) ? cur.state ^ EM : cur.state;
		    cur.s = "<emphasis>";
		    this = startnode('E',this,&cur);
		}
	    } else {
		fprintf(stderr, "malformed emphasis at position %d\n", i);
		return 0;
	    }    
	    break;
	  case '0':
	  case '1':
	  case '2':
	  case '3':
	  case '4':
	  case '5':
	  case '6':
	  case '7':
	  case '8':
	  case '9':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
		if(inclose) {
		    if(verbose) printf("</size>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.size = frame.size;
		} else {
		    if(verbose) printf("<size>");
		    push(&cur);
		    cur.size = p[i] - '0';
		    cur.s = "<size>";
		    this = startnode(p[i],this,&cur);
		}
	    }
	    break;
	  case '/':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag==1)
		if(inclose) {
		    fprintf(stderr, "malformed /-tag at position %d\n", i);
		    return 0;
		} else {
		    inclose = 1;
		}
	    break;
	  case 'r':
	  case 'g':
	  case 'b':
	  case 'c':
	  case 'm':
	  case 'y':
	  case 'k':
	  case 'w':
	    if(!intag)
		emit(p[i],&cur);
	    else if(intag == 1) {
		intag=2;
		if(inclose) {
		    if(verbose) printf("</color>");
		    this = finishnode(this);
		    pop(&frame);
		    cur.color = frame.color;
		} else {
		    if(verbose) printf("<color>");
		    push(&cur);
		    cur.color = p[i];
		    cur.s = "<color>";
		    this = startnode(p[i],this,&cur);
		}
	    } else {
		fprintf(stderr, "malformed color at position %d\n", i);
		return 0;
	    }
	    break;

#define ISSP(c) ((c) == ' ' || (c) == '\t' || (c) == 0xa || (c) == 0xd)
#define SPC ' '

	  case ' ':
	  case '\t':
	  case 0xd:
	  case 0xa:
	    space = cur;
	    space.state &= ~(S | EM | I | B);
	    if(!(space.state & U))
		space.color = 'w';
	    if(space.state & TT) {
		emit(SPC, &space);
	    } else if(ISSP(lastemit) &&
		      last_space.state == space.state &&
		      eq(last_space.size,space.size) &&
		      eq(last_space.color,space.color)) {
	    } else {
		emit(SPC,&space);
	    }
	    last_space = space;
	    break;

	  default:
	    emit(p[i],&cur);
	    break;
	}
    }
    finishnode(this);
    return 1;
}

int main(int argc, char *argv[]) {
    if(argc != 1) {
	fprintf(stderr, "usage: %s <html-infile\n", argv[0]);
	exit(1);
    }
    init();
    if(!readfile(0)) {
	fprintf(stderr, "%s: read failed", argv[0]);
	exit(1);
    }
    contextualize(f, f_sz);
    if(0) {
	dumpparse(0);
	putchar('\n');
    }
    if(!emitdecor) {
	int i;
	revparse(0);
	if(0)
	for(i=0; i < 5; i++) {
	    optCoalese(0);
	    optDead(0);
	    optSEM(0,0);
	    optEM(0);
	}
	emitparse(0,0);
//	dumpparse(0);
    }
}
