(* -*- sml -*-
 * factor.cpp
 *
 * (C) 2001 Lucent Technologies, Bell Labs
 *)
structure Factor : sig
    val factor : Document.tagging * Document.tagging ->
		 { common: Document.tagging,
		   left: Document.tagging,
		   right: Document.tagging }
end = struct

    structure D = Document

#include "bits.ins"

    infix || &&
    val op || = Word.orb
    val op && = Word.andb

    val bittsem_bits = b_bit || i_bit || tt_bit || s_bit || em_bit

    fun factor (t1: D.tagging, t2: D.tagging) = let
	fun colsz sel = let val v1 = sel t1 and v2 = sel t2
			in if v1 = v2 then (v1, ~1, ~1) else (~1, v1, v2)
			end
	val (ccolor, rcolor1, rcolor2) = colsz #color
	val (csz, rsz1, rsz2) = colsz #sz
	val font1 = #font t1
	val font2 = #font t2
	val pl1 = testBit pl_bit font1
	val pl2 = testBit pl_bit font2
	val (cfont, rfont1, rfont2) =
	    if pl1 = pl2 then
		let val pl = pl1
		    val u1 = projBits u_bits font1
		    val u2 = projBits u_bits font2
		    val cu = Word.min (u1, u2)
		    val ru1 = u1 - cu
		    val ru2 = u2 - cu
		    val bittsem1 = font1 && bittsem_bits
		    val bittsem2 = font2 && bittsem_bits
		    val cbittsem = bittsem1 && bittsem2
		    val not_cbittsem = Word.notb cbittsem
		    val rbittsem1 = bittsem1 && not_cbittsem
		    val rbittsem2 = bittsem2 && not_cbittsem
		    val cpl = if pl then pl_bit else 0w0
		in
		    (cpl || cbittsem || cu, rbittsem1 || ru1, rbittsem2 || ru2)
		end
	    else (0w0, font1, font2)
    in
	{ common = { color = ccolor, sz = csz, font = cfont },
	  left = { color = rcolor1, sz = rsz1, font = rfont1 },
	  right = { color = rcolor2, sz = rsz2, font = rfont2 } }
    end

end
