(* -*- sml -*-
 * fastiinfer.sml
 *
 * (C) 2001 Lucent Technologies, Bell Labs
 *)
structure FastIInfer : sig

    (* ignore info that two documents have in common *)
    val i_pair : Document.ignore * Document.ignore -> Document.ignore

    (* ignore nothing *)
    val i_none : Document.ignore

    (* ignore everything *)
    val i_everything : Document.ignore

    (* fold i_pair over a list *)
    val i_list : Document.ignore list -> Document.ignore

    (* recalculate ignore info after more tags are being added *)
    val i_elem : Document.tagging * Document.ignore -> Document.ignore

    (* strip tags that are ignored *)
    val i_strip : Document.tagging * Document.ignore -> Document.tagging

    (* strip tags that are NOT ignored *)
    val i_safe : Document.tagging * Document.ignore -> Document.tagging

end = struct

#include "sub.ins"

    local
	structure D = Document
	structure SII = SlowIInfer
    in

    val i_pair = SII.i_pair		(* for now *)

    val i_none = SII.i_none
    val i_everything = SII.i_everything

    fun i_list [] = i_everything
      | i_list (h :: t) = foldl i_pair h t

    fun i_elem (t: D.tagging, i:D.ignore) = let
	val fast_font = stringSubW(arraySubW(Tables.ignoreElemTbl,
					     #font t),
				   #font i)
#ifndef NDEBUG
	val slow_font = #font (SII.i_elem (t, i))
	val _ = if fast_font <> slow_font then
		    (TextIO.output (TextIO.stdErr,
				    "!!! i_elem discrepancy!\n");
		     OS.Process.exit OS.Process.failure)
		else ()
#endif
    in
	{ color = #color t <> ~1 orelse #color i,
	  sz = #sz t <> ~1 orelse #sz i,
	  font = fast_font }
    end

    val i_strip = SII.i_strip		(* for now *)

    val i_safe = SII.i_safe		(* for now *)

    end (* local *)
end
