(* main.sml
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *)

structure Main : sig

    val main : (string * string list) -> OS.Process.status

  end = struct

    val limit = ref(Time.fromSeconds 180)
    val outputTime = Time.fromSeconds 5

    fun setTimeLimit t = (case Time.fromString t
	   of NONE => ()
	    | SOME t => limit := t
	  (* end case *))

    fun doOpts [] = (case OS.Process.getEnv "TLIMIT"
	   of NONE => ()
	    | SOME t => setTimeLimit t
	  (* end case *))
      | doOpts (t::r) = setTimeLimit t

    fun errMsg msg = TextIO.output (TextIO.stdErr, concat msg)

    fun applyOpts d optl = 
	foldl (fn (opt,d) => opt d) d optl

    val basicOpts = [
	    EMOpt.emopt, TDContract.contract,
	    (* Merge.optPL, TDContract.contract, *)  (* broken *)
	    Merge.merge, TDContract.contract
	  ]
    val dpOpt = [Nesting.doit 8,
		 Nesting.doit 16,
		 Nesting.doit 32]

    fun main (cmd, argv) = let
	  val t0 = Time.now()
	  val _ = doOpts argv;
	  val doc = applyOpts (Input.input()) basicOpts
	  val elapsed = Time.-(Time.now(), t0) handle _ => Time.zeroTime
	  val remaining = Time.-(!limit, elapsed)
	  val doc = (
		  TimeLimit.timeLimit (Time.-(remaining, outputTime))
		    (applyOpts doc) dpOpt
		) handle _ => (
		  TextIO.output(TextIO.stdErr, "!!!running out of time!!!\n");
		  doc)
	  in
	    Output.output doc;
	    OS.Process.success
	  end handle exn => (
	    errMsg [
		"uncaught exception ", exnName exn, " [", exnMessage exn, "]\n"
	      ];
	    app (fn s => errMsg ["  raised at ", s, "\n"]) (SMLofNJ.exnHistory exn);
	    OS.Process.failure)

  end
