(* -*- sml -*-
 * merge.cpp
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *
 * simple merge optimization
 *)

structure Merge : sig

    val merge : Document.doc -> Document.doc
    val optSEM : Document.doc -> Document.doc
    val optPL : Document.doc -> Document.doc
  end = struct

    structure D = Document
    structure SS = Substring
    structure U = Unsafe

#include "bits.ins"


  (* subscript operations *)
(*
    val arraySub = Array.sub
    val stringSub = String.sub
*)
    val arraySub = U.Array.sub
    val stringSub = U.CharVector.sub
    fun arraySubW (a, w : word) = arraySub(a, Word.toIntX w)
    fun stringSubW (s, w : word) = Word.fromInt(ord(stringSub(s, Word.toIntX w)))
    fun t_col (tg: D.tagging) = #color tg
    fun t_sz (tg: D.tagging) = #sz tg
    fun t_f (tg: D.tagging) = #font tg

    fun muopen t = 
	let val n = if t_col t = ~1 then 3 else 0
            val n = n+(if t_sz t = ~1 then 3 else 0)
	    val f = t_f t
            val n = n+(if Word.andb(f,i_bit) <> 0w0 then 3 else 0)
            val n = n+(if Word.andb(f,b_bit) <> 0w0 then 3 else 0)
            val n = n+(if Word.andb(f,tt_bit) <> 0w0 then 4 else 0)
            val n = n+(if Word.andb(f,em_bit) <> 0w0 then 4 else 0)
            val n = n+(if Word.andb(f,s_bit) <> 0w0 then 3 else 0)
            val n = n+(if Word.andb(f,pl_bit) <> 0w0 then 4 else 0)
            val n = n+((Word.toInt (Word.andb(f,u_bits))) * 3)
	in	
		n
	end

    fun muclose t = 
	let val n = if t_col t = ~1 then 4 else 0
            val n = n+(if t_sz t = ~1 then 4 else 0)
	    val f = t_f t
            val n = n+(if Word.andb(f,i_bit) <> 0w0 then 4 else 0)
            val n = n+(if Word.andb(f,b_bit) <> 0w0 then 4 else 0)
            val n = n+(if Word.andb(f,tt_bit) <> 0w0 then 5 else 0)
            val n = n+(if Word.andb(f,em_bit) <> 0w0 then 5 else 0)
            val n = n+(if Word.andb(f,s_bit) <> 0w0 then 4 else 0)
            val n = n+(if Word.andb(f,pl_bit) <> 0w0 then 5 else 0)
            val n = n+((Word.toInt (Word.andb(f,u_bits))) * 4)
	in	
		n
	end



    fun bits(w: word) = 
	let fun aux (w,n) = if(w = 0wx0) then n
			    else if(Word.andb(w,0wx1) = 0wx1) then
				aux(Word.>>(w,0wx1),n+1)
			    else
				aux(Word.>>(w,0wx1),n)
	in aux(w,0)
	end

    fun bits0(w: word) =  (* number of bits for flags + u value *)
	let fun aux (w,n) = if(w = 0wx0) then n
			    else if(Word.andb(w,0wx1) = 0wx1) then
				aux(Word.>>(w,0wx1),n+1)
			    else
				aux(Word.>>(w,0wx1),n)
	in aux(Word.>>(w,0w2),Word.toInt (Word.andb(u_bits,w)))
	end

    fun mergeContent (D.LIST{body=b1,ign=i1},
		      D.LIST{body=b2,ign=i2}) =
		D.LIST{body=b1@b2,ign=i1}
      | mergeContent (D.LIST{body=b,ign=i},x) = 
		D.LIST{body=b@[x],ign=i}
      | mergeContent (x,D.LIST{body=b,ign=i}) = 
		D.LIST{body=x::b,ign=i}
      | mergeContent (x,y) = 
		D.LIST{body=[x,y],ign=D.ignOf x}

    infix ||
    val op || = Word.orb

    fun	dwnCtxt (tags:D.tagging,ctxt:D.tagging) : D.tagging = 
	let val cf = t_f ctxt
	    val tf = t_f tags
	    val col = if t_col ctxt = ~1 then t_col tags
		      else if t_col tags = ~1 then t_col ctxt
		      else t_col tags
	    val size = if t_sz ctxt = ~1 then t_sz tags
		      else if t_sz tags = ~1 then t_sz ctxt
		      else t_sz tags
	in
		if Word.andb(tf,pl_bit) <> 0w0 then 
			{color=col, sz=size, font=0w0}
		else
		let val cu = Word.andb(u_bits,cf)
	    	    val tu = Word.andb(u_bits,tf)
	    	    val u = cu+tu
	    	    val u = if u>0w3 then 0w3 else u
		    val b = Word.orb(Word.andb(tf,b_bit),
				     Word.andb(cf,b_bit))
		    val i = Word.orb(Word.andb(tf,i_bit),
				     Word.andb(cf,i_bit))
		    val tt = Word.orb(Word.andb(tf,tt_bit),
				      Word.andb(cf,tt_bit))
		    val s = Word.orb(Word.andb(tf,s_bit),
				     Word.andb(cf,s_bit))
		    val em = if s = 0w0 then
				Word.orb(Word.andb(tf,em_bit),
				         Word.andb(cf,em_bit))
			     else
				0w0
		   val fnt = u || b || i || tt || s || em

		in
			{color=col,sz=size,font=fnt}
		end
	end
	    
    fun pr (D.ELEM _) = print "ELEM "
      | pr (D.TEXT _) = print "TEXT "
      | pr (D.SP _) = print "SP "
      | pr (D.LIST _) = print "LIST "

    fun doVar (e1 as D.ELEM{tags=t1,content=c1,ign=i1},
	       ez as D.ELEM{tags=tz,content=cz,ign=iz},
	       stuff, ctxt) =
		let val t1b = t_f t1
		    val t1bn = bits0 t1b
		    val t2b = t_f ctxt
		    val keepcol = t_col t1 <> t_col ctxt
		    val keepsz = t_sz t1 <> t_sz ctxt
		    val t2bn = (bits0 t2b)

(*
		val _ = print ("t1bn is "^(Int.toString t1bn))
		val _ = print ("t2bn is "^(Int.toString t2bn))
*)


		    val save = (muclose t1) + (muopen t1) + (if keepcol then 8 else 0) + (if keepcol then 8 else 0)
		    val cost = (muopen ctxt) + (muclose ctxt) + 9
(*
		val _ = print (" save is "^(Int.toString save))
		val _ = print (" cost is "^(Int.toString cost))
*)
		    val heavy = save > cost
(*
		    val heavy = t1bn*2 + t2bn >
				1*(t1bn + (t2bn+1) +
			       (if keepsz then 1 else 0) +
			       (if keepcol then 1 else 0))
*)

		in
			if(heavy) then
				let (* val _ = print "found heavy opp\n" *)
				    val c1 = if keepcol then
						D.ELEM{tags={color=t_col t1,
							     sz= ~1,font=0w0},
						       content=c1,
						       ign=i1}
					     else c1

				    val cz = if keepcol then
						D.ELEM{tags={color=t_col t1,
							     sz= ~1,font=0w0},
						       content=cz,
						       ign=i1}
					     else cz

				    val c1 = if keepsz then
						D.ELEM{tags={color= ~1,
							     sz= t_sz t1,font=0w0},
						       content=c1,
						       ign=i1}
					     else c1

				    val cz = if keepsz then
						D.ELEM{tags={color= ~1,
							     sz= t_sz t1,font=0w0},
						       content=cz,
						       ign=i1}
					     else cz

				    val t1 = {color=if keepcol then ~1 else t_col t1,
	                                      sz=if keepsz then ~1 else t_sz t1,
					      font=t_f t1}

				    val mid = D.ELEM{tags={color= ~1,sz= ~1, font=pl_bit},
						     content=D.ELEM{tags=ctxt,
						     		    content=D.LIST{body=stuff,ign=i1},
						     		    ign=i1},
						     ign=i1}

				    val m = mergeContent(mergeContent(c1,mid),cz)
				in 
					[D.ELEM{tags=t1,content=m,ign=i1 (*??*)}]
			        end
			else
				e1 :: (stuff@[ez])
		end
	| doVar (e,_,_,_) = [e]

    fun candidate (l,t) =
	let fun aux((D.SP _)::r) = aux r
              | aux((D.TEXT _)::r) = aux r
	      | aux((D.ELEM{tags=s,...})::_) = s = t
              | aux _ = false
	in aux l
	end

    fun getParts l = 
	let fun aux((x as (D.SP _))::r,stuff) = aux (r,x::stuff)
              | aux((x as (D.TEXT _))::r,stuff) = aux (r,x::stuff)
	      | aux((x as (D.ELEM _))::r,stuff) = (rev stuff, x, r)
	in aux (l,[])
	end

    fun optPLaux (l as ((e1 as D.ELEM{tags=t1,content=c1,ign=i1})::
	          (e2 as D.ELEM{tags=t2,content=c2,ign=i2})::
	          (e3 as D.ELEM{tags=t3,content=c3,ign=i3})::rest), ctxt) =
	let val newhd = 
		if t1 = t3 then
		let val t1b = t_f t1
		    val t1bn = bits0 t1b
		   (* bits needed to restore context and correctly mark e2 *)
		    val nctxt = dwnCtxt(t2,ctxt)
		    val t2b = t_f nctxt
		    val keepcol = t_col t1 <> t_col nctxt andalso t_col t2 = ~1
		    val keepsz = t_sz t1 <> t_sz nctxt andalso t_sz t2 = ~1
		    val t2bn = (bits0 t2b)
	
		    val save = (muclose t1) + (muopen t1) + (if keepcol then 8 else 0) + (if keepcol then 8 else 0)
		    val cost = (muopen ctxt) + (muclose ctxt) + 9
(*
		val _ = print (" save is "^(Int.toString save))
		val _ = print (" cost is "^(Int.toString cost))
*)
		    val heavy = save > cost

(*
		    val heavy = t1bn*2 + t2bn >
				t1bn + (t2bn+1) +
			       (if keepsz then 1 else 0) +
			       (if keepcol then 1 else 0)
*)
		in
			if(heavy) then
				let (* val _ = print "found heavy opp\n" *)
				    val tx = {color=t_col nctxt,
					      sz=t_sz nctxt,
					      font=t2b}
				    val midx = D.ELEM{tags=tx,
						      content=c2,
						      ign=i2}
				    val ty = {color=t_col nctxt,
					     sz=t_sz nctxt,
					     font=pl_bit}
				    val midy = D.ELEM{tags=ty,
						      content=midx,
						      ign=i2 (*??*)}
				    val c1 = if keepcol then
						D.ELEM{tags={color=t_col t1,
							     sz= ~1,font=0w0},
						       content=c1,
						       ign=i1}
					     else c1

				    val c3 = if keepcol then
						D.ELEM{tags={color=t_col t1,
							     sz= ~1,font=0w0},
						       content=c3,
						       ign=i1}
					     else c3

				    val c1 = if keepsz then
						D.ELEM{tags={color= ~1,
							     sz= t_sz t1,font=0w0},
						       content=c1,
						       ign=i1}
					     else c1

				    val c3 = if keepsz then
						D.ELEM{tags={color= ~1,
							     sz= t_sz t1,font=0w0},
						       content=c3,
						       ign=i1}
					     else c3

				    val t1 = {color=if keepcol then ~1 else t_col t1,
	                                      sz=if keepsz then ~1 else t_sz t1,
					      font=t_f t1}


				    val m = mergeContent(mergeContent(c1,midy),c3)
				in 
					SOME(D.ELEM{tags=t1,
					            content=m,ign=i1 (*??*)})
			        end
			else NONE
		end
		else NONE
	    val l = case newhd of
			NONE => l
		      | SOME x => x::rest
	in
		(optPL (hd l,ctxt))::(optPLaux(tl l,ctxt))
	end
      | optPLaux ((l as ((e1 as D.ELEM{tags=t1,content=c1,ign=i1})::
	          l2)), ctxt) =
		let fun f () = (optPL (e1,ctxt))::(optPLaux (l2,ctxt))
		in
			if candidate (l2,t1) then
				let (* val _ = print "found a candidate" *)
				    val (stuff, ez, rest) = getParts l2
				    val newhd = doVar(e1, ez, stuff, ctxt)
				    val newl = newhd@rest
				in
					(optPL((hd newl),ctxt))::(optPLaux(tl newl,ctxt))
				end
			else f()
		end
      | optPLaux (x::l,ctxt) = (optPL (x,ctxt))::(optPLaux (l,ctxt))
      | optPLaux ([],_) = []
    and
        optPL (D.LIST{body=b,ign=i},ctxt) = 
		D.LIST{body=optPLaux (b,ctxt),ign=i}
      | optPL (D.ELEM{tags=t,content=c,ign=i},ctxt) =
	let 	val nctxt = dwnCtxt(t,ctxt)
	in
		D.ELEM{tags=t,content=optPL (c,nctxt),ign=i}
	end
      | optPL (x,_) = x

    val ctxt1 = {color= ~1, sz= ~1, font=0w0}
    val optPL = fn d => optPL(d, ctxt1)

    fun optSEM (D.LIST{body=b,ign=i},seen) =
	D.LIST{body=optSEMl (b,seen),ign=i}
      | optSEM (D.ELEM{tags=t,content=c,ign=i},seen) = 
	let val now = seen orelse (testBit s_bit (t_f t))
	    val t0 = if now then 
			let val nbit = clearBit em_bit (t_f t)
			    val t1 = {color=t_col t,sz=t_sz t,font=nbit}
			in t1 end
		     else t
	    val c0 = optSEM(c,now)
	in
		D.ELEM{tags=t0,content=c0,ign=i}
	end
      | optSEM (x,_) = x
    and
        optSEMl (x::l,seen) = (optSEM(x,seen))::(optSEMl(l,seen))
      | optSEMl ([],_) = []

    val optSEM = fn x => optSEM(x,false)

    fun aux ((e as D.ELEM{tags=t1,content=c1,ign=i1})::
	     (rest as (D.ELEM{tags=t2,content=c2,ign=i2}::l))) = 
		if(t1=t2) then
			aux(D.ELEM{tags=t1,content=mergeContent(c1,c2),ign=i1}::l)
		else
			(merge e)::(aux rest)
      | aux (x::l) = (merge x)::(aux l)
      | aux [] = []
    and
        merge (D.LIST{body=b,ign=i}) = 
		D.LIST{body=aux b,ign=i}
      | merge (D.ELEM{tags=t,content=c,ign=i}) =
		D.ELEM{tags=t,content=merge c,ign=i}

      | merge x = x
	

  end
