(* document.sml
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *)

structure Document =
  struct

    datatype tag
      = B | EM | I | PL | S | TT | U
      | Sz of int
      | Col of int

    fun tagToString B = "B"
      | tagToString EM = "EM"
      | tagToString I = "I"
      | tagToString PL = "PL"
      | tagToString S = "S"
      | tagToString TT = "TT"
      | tagToString U = "U"
      | tagToString (Sz n) = Int.toString n
      | tagToString (Col c) = str(String.sub("rgbcmykw", c))

  (* The following type describes a "canonical" tagging annotation.
   * The idea is that successive applications of such taggings can
   * always be collapsed into a single one.
   * Invariant: s implies not em.
   *)
    type tagging = {
	color : int,		(* ~1 means don't care *)
	sz : int,		(* ~1 means don't care *)
	pl : bool,
	u : int,
	b : bool,
	i : bool,
	tt : bool,
	s : bool,
	em : bool
      }

  (* A document "ignores" or "tolerates" a tag if no matter
   * whether or not that tag is added as an outside wrapper
   * it won't change the meaning of the document.
   *)
    type ignored = {
	  i_color : bool,
	  i_sz : bool,
	  i_pl : bool,
	  i_u : int,			(* ignores all beyond that many *)
	  i_b : bool,
	  i_i : bool,
	  i_tt : bool,
	  i_s : bool,
	  i_em : bool
	}

(*
    datatype doc
      = SP of ignored
      | TEXT of (Substring.substring * ignored)
      | DOC of {
	  tags : tagging,
	  ignores : ignored,
	  length : int,
	  content : doc list
	}
*)

    datatype doc
      = SP
      | TEXT of Substring.substring
      | DOC of {
	  tag : tag,
	  content : doc list
	}

  end
