
structure Gen = struct

  structure R = Random
  structure D = Document
  structure SS = Substring

  val seed = ref (R.rand (0,0))

  val maxDepth = ref (5)

  val depth = ref (0)

  fun newSeed (n1) = (seed := (R.rand (n1,0)))

  val count = ref 0
  fun docount _ = let val x = !count
	          in count := !count+1;
	             x
		  end

  (* return a random between 1 and n, inclusive *)
  fun randMax (n) = R.randRange (1,n) (!seed)
  
  (* for now, everything equally likely *)
  fun randomText () = 
    case (randMax (3)) 
      of 1 => D.TEXT (SS.all "  ")
       | 2 => D.TEXT (SS.all ("foo"^Int.toString(docount())))
       | 3 => D.TEXT (SS.all ("  foo"^Int.toString(docount())^"  "))

  fun randomTag () = let
    val tag = case (randMax (25))
            of 1 => D.B
             | 2 => D.EM
             | 3 => D.I
             | 4 => D.PL
             | 5 => D.S
             | 6 => D.TT
             | 7 => D.U
             | 8 => D.Col (0)
             | 9 => D.Col (1)
             | 10 => D.Col (2)
             | 11 => D.Col (3)
             | 12 => D.Col (4)
             | 13 => D.Col (5)
             | 14 => D.Col (6)
             | 15 => D.Col (7)
             | n => D.Sz (n-16)
  in
    D.DOC {tag=tag, content = randomDocList ()}
  end

  and randomDoc () = 
    case (randMax (2))
      of 1 => randomText ()
       | 2 => randomTag ()

  and randomDocList () = let
    val l = List.tabulate (randMax (10), fn i => ())
  in
    if (!depth > !maxDepth)
      then [randomText ()]
    else let
      val _ = depth := (!depth)+1
      val r = map (fn _ => randomDoc ()) l
    in
      depth := (!depth)-1;
      r
    end
  end
    
  fun gen () = let
    val _ = newSeed (Int32.toInt (Time.toSeconds (Time.now ())))
  in
    randomDocList ()
  end

end
