(* output.sml
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *
 * Output routine for SML/NG documents.  This prints the representation produced
 * by the input routine.  The optimization takes place on the document structure,
 * without touching the underlying text (the string list).
 *)

structure Output : sig

    val output : Document.doc list -> unit

  end = struct

    structure D = Document
    structure SIO = TextIO.StreamIO
    structure SS = Substring

    fun output doc = let
	  val outStrm = TextIO.getOutstream TextIO.stdOut
	  fun outText ss = SIO.output(outStrm, SS.string ss)
	(* we print spaces as newlines for the top-level content, since it makes
	 * the output a bit easier to read.
	 *)
	  fun topOut [] = ()
	    | topOut (D.SP::r) = (SIO.output1(outStrm, #"\n"); topOut r)
	    | topOut ((D.TEXT ss)::r) = (outText ss; topOut r)
	    | topOut (D.DOC{tag, content}::r) = (outDoc(tag, content); topOut r)
	  and outDoc (tag, content) = (
		SIO.output(outStrm, String.concat["<", D.tagToString tag, ">"]);
		out content;
		SIO.output(outStrm, String.concat["</", D.tagToString tag, ">"]))
	  and out [] = ()
	    | out (D.SP::r) = (SIO.output1(outStrm, #" "); out r)
	    | out ((D.TEXT ss)::r) = (outText ss; out r)
	    | out (D.DOC{tag, content}::r) = (outDoc(tag, content); out r)
	  in
	    ignore (topOut doc)
	  end

  end
