(* -*- sml -*-
 * slowiinfer.sml
 *
 * (C) 2001 Lucent Technologies, Bell Labs
 *)

structure SlowIInfer : sig

    (* ignore info that two documents have in common *)
    val i_pair : Document.ignore * Document.ignore -> Document.ignore

    (* ignore nothing *)
    val i_none : Document.ignore

    (* ignore everything *)
    val i_everything : Document.ignore

    (* fold i_pair over a list *)
    val i_list : Document.ignore list -> Document.ignore

    (* recalculate ignore info after more tags are being added *)
    val i_elem : Document.tagging * Document.ignore -> Document.ignore

    (* strip tags that are ignored *)
    val i_strip : Document.tagging * Document.ignore -> Document.tagging

    (* strip tags that are NOT ignored *)
    val i_safe : Document.tagging * Document.ignore -> Document.tagging

end = struct

    structure D = Document

#include "bits.ins"

    infix && ||
    val op && = Word.andb
    val op || = Word.orb

    fun i_pair (i1: D.ignore, i2: D.ignore) = let
	val f1 = #font  i1
	val f2 = #font  i2

	fun both sel = sel i1 andalso sel i2

	val f = (f1 && f2 && Word.notb u_bits) ||
                Word.max (projBits u_bits f1, projBits u_bits f2)
    in
	{ color = both #color, sz = both #sz, font = f } : D.ignore
    end

    (* all bits off except for u: u = 3 *)
    val i_none = { color = false, sz = false, font = 0w3 }

    (* all bits are on except for u: u = 0 *)
    val i_everything = { color = true, sz = true,
			 font = em_bit || s_bit || tt_bit || i_bit ||
				b_bit || pl_bit }

    fun i_list [] = i_everything
      | i_list (h :: t) = foldl i_pair h t

    fun i_elem (t: D.tagging, ig: D.ignore) = let
    	val ft = #font t
	val tpl = testBit pl_bit ft
	val fi = #font ig

	val u = if tpl then 0w0
		else Word.max (projBits u_bits fi - projBits u_bits ft, 0w0)
	val b = tpl orelse testBit b_bit ft orelse testBit b_bit fi
	val i = tpl orelse testBit i_bit ft orelse testBit i_bit fi
	val tt = tpl orelse testBit tt_bit ft orelse testBit tt_bit fi
	val s = tpl orelse testBit s_bit ft orelse testBit s_bit fi
	val em = tpl orelse s orelse testBit em_bit fi

	val pl = tpl orelse testBit pl_bit fi

	val col = #color t <> ~1 orelse #color ig
	val sz = #sz t <> ~1 orelse #sz ig

	fun bb _ false = 0w0
	  | bb b true = b
        val f = bb pl_bit pl || bb b_bit b || bb i_bit i || bb tt_bit tt ||
                bb s_bit s || bb em_bit em || u
    in
	{ color = col, sz = sz, font = f } : D.ignore
    end

    fun i_strip (t: D.tagging, ig: D.ignore) = let
	val ft = #font t
	val fi = #font ig

	val ft_nou = ft && Word.notb u_bits
	val fi_nou = fi && Word.notb u_bits

	val ut = projBits u_bits ft
	val ui = projBits u_bits fi
	val u = Word.min (ut, ui)

	val f_nou = ft_nou && Word.notb fi_nou
    in
	{ color = if #color ig then ~1 else #color t,
	  sz = if #sz ig then ~1 else #sz t,
	  font = f_nou || u }
    end

    fun i_safe (t: D.tagging, ig: D.ignore) = let

	val ft = #font t
	val fi = #font ig

	val ft_nou = ft && Word.notb u_bits
	val fi_nou = fi && Word.notb u_bits

	val ut = projBits u_bits ft
	val ui = projBits u_bits fi
	val u = if ut = 0w0 then ut else 0w0

	val f_nou = ft_nou && fi_nou
    in
	{ color = if #color ig then #color t else ~1,
	  sz = if #sz ig then #sz t else ~1,
	  font = f_nou || u }
    end
end
