(* -*- sml -*-
 * slowtagging.cpp
 *
 * (C) 2001 Lucent Technologies, Bell Labs
 *)
structure SlowTagging = struct

#include "bits.ins"

    fun single_color c = { color = c, sz = ~1, font = 0w0 }
    fun single_sz s = { color = ~1, sz = s, font = 0w0 }
    fun single_font f =
	{ color = ~1,
	  sz = ~1,
	  font = case f of
		     6 => pl_bit
		   | 5 => b_bit
		   | 4 => i_bit
		   | 3 => tt_bit
		   | 2 => s_bit
		   | 1 => em_bit
		   | _ => 0w1		(* u = 1 *) }

    structure D = Document

    fun tcollapse (t1: D.tagging, t2: D.tagging) : D.tagging = let
	fun prefer2 sel = case sel t2 of ~1 => sel t1 | x => x

	val color = prefer2 #color
	val sz = prefer2 #sz

	val f1 = #font t1
	val f2 = #font t2

	fun either sel = sel f1 orelse sel f2

	val f_pl = testBit pl_bit
	val f_b = testBit b_bit
	val f_i = testBit i_bit
	val f_tt = testBit tt_bit
	val f_s = testBit s_bit
	val f_em = testBit em_bit
	val f_u = projBits u_bits

	val (pl, u, b, i, tt, s, em) =
	    if f_pl f2 then
		(true, f_u f2, f_b f2, f_i f2, f_tt f2, f_s f2, f_em f2)
	    else let val u0 = f_u f1 + f_u f2
		     val usat = u0 >= 0w3(* saturated u *)
		     val u = if usat then 0w3 else u0
		     val b = either f_b
		     val i = either f_i
		     val tt = either f_tt
		     val s = either f_s
		     val em = not s andalso (f_em f1 <> f_em f2)
		     (* PL will not matter if U is already saturated and
		      * B/I/TT are all set and S is set.
		      * (Just EM instead of S does not cut it because
		      * PL can erase an outside S.  S means "strength"
		      * is saturated, EM does not mean that.) *)
		     val pl = 
			 not (usat andalso b andalso i andalso tt andalso s)
			 andalso f_pl f1
		 in
		     (pl, u, b, i, tt, s, em)
		 end

	fun bb _ false = 0w0
	  | bb b true = b
	infix ||
	val op || = Word.orb
        val f = bb pl_bit pl || bb b_bit b || bb i_bit i || bb tt_bit tt ||
                bb s_bit s || bb em_bit em || u
    in
	{ color = color, sz = sz, font = f }
    end
end
